# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2018, 2019. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

from pyspark.sql import SparkSession
from snap_ml_spark import DatasetReader
import numpy as np

class SupportVectorMachine:
    """
    Support Vector Machine classifier
    
    This class implements regularized support vector machine using the IBM Snap ML solver. It can handle sparse and dense dataset formats. Use libsvm, snap or csv format for the Dual algorithm, or snap.t (transposed) format for the primal algorithm.

    Parameters
    ----------
    max_iter : int, default : 1000
        Maximum number of iterations used by the solver to converge.

    regularizer : float, default : 1.0
        Regularization strength. It must be a positive float.
        Larger regularization values imply stronger regularization.

    verbose: boolean, default : False
        Flag for indicating if the training loss will be printed at each epoch.

    use_gpu : bool, default : False
        Flag for indicating the hardware platform used for training. If True, the training
        is performed using the GPU. If False, the training is performed using the CPU.

    class_weights : 'balanced'/True or None/False, optional
        If set to 'None', all classes will have weight 1.
    
    gpu_mem_limit : int, default : 0
        Limit of the GPU memory. If set to the default value 0, 
        the maximum possible memory is used.

    n_threads : int, default : -1 meaning that n_threads=256 if GPU is enabled, else 1
        Number of threads to be used.

    tol : float, default : 0.001
        The tolerance parameter. Training will finish when maximum change in model coefficients is less than tol. 

    return_training_history : str or None, default : None
        How much information about the training should be collected and returned by the fit function. By 
        default no information is returned (None), but this parameter can be set to "summary", to obtain 
        summary statistics at the end of training, or "full" to obtain a complete set of statistics 
        for the entire training procedure. Note, enabling either option will result in slower training.

    labelColIndex : int, default : 1
        Denotes which column to be marked as label.
        It's applicable only when Dataframe is passed in as input and its holding features in individual columns
        rather than in DenseVector or SparseVector

    Attributes
    ----------
    coef_ : ndarray, shape (n_features,)
        Coefficients of the features in the trained model.

    pred_array_: ndarray, shape(number_of_test_examples,)
        binary predictions written by the predict() function
        
    """

    def __init__(self, max_iter=1000, regularizer=1.0, verbose=False, use_gpu=False, class_weights=None, gpu_mem_limit=0, n_threads=-1, tol=0.001, return_training_history=None, labelColIndex=1):
        
        ss = SparkSession.builder.getOrCreate()
        
        # Create an instance of the DatasetReader using the JVM from the SparkSession 
        self._SupportVectorMachine = ss._jvm.com.ibm.snap.ml.SupportVectorMachine()
    
        self.__params = {}

        self.pred_array_ = []
        self.coef_ = None

        # Set GPU memory limit
        if gpu_mem_limit >= 0:
            self.__params['gpu_mem_limit'] = gpu_mem_limit
            self._SupportVectorMachine = self._SupportVectorMachine.setGpuMemLimit(gpu_mem_limit)
        if gpu_mem_limit < 0 and use_gpu==True:
            print('WARNING: Invalid gpu_mem_limit value, set back to 0')
            self.__params['gpu_mem_limit'] = 0
            self._SupportVectorMachine = self._SupportVectorMachine.setGpuMemLimit(0)

        # set label column index
        if labelColIndex > 0:
            self.__params['labelColIndex'] = labelColIndex
        else:
            raise ValueError("Invalid labelColIndex value")

        # Set the number of threads
        if n_threads == -1:
            if use_gpu == True:
                n_threads = 256
            else:
                n_threads = 1
        else:
            if use_gpu:
                if n_threads < 32:
                    raise ValueError("When using GPU, n_threads must be >= 32")
                if n_threads > 1024:
                    raise ValueError("When using GPU, n_threads must be <= 1024")
            else:
                if n_threads < 1:
                    raise ValueError("When using CPU, n_threads must be >= 1")

        self.__params['n_threads'] = n_threads
        self._SupportVectorMachine = self._SupportVectorMachine.setNThreads(n_threads)

        # Set if the class weights will be used.
        if class_weights in set([None,False]):
            self.__params['class_weights'] = False
            self._SupportVectorMachine = self._SupportVectorMachine.setBalanced(False)
        elif class_weights in set(['balanced',True]):
            self.__params['class_weights'] = True
            self._SupportVectorMachine = self._SupportVectorMachine.setBalanced(True)
        else:
            raise ValueError("Invalid class_weights value: class_weight must be set to None/False or 'balanced'/True")

        # Set the dual
        self._SupportVectorMachine = self._SupportVectorMachine.setDual(True)

        # Set maximum number of epochs to be performed by the distributed algorithm
        if max_iter > 0:
            self.__params['max_iter'] = max_iter
            self._SupportVectorMachine = self._SupportVectorMachine.setMaxIter(max_iter)
        else:
            raise ValueError("Invalid max_iter value")

        # Set regularizer
        if regularizer >=0:
            self.__params['regularizer'] = regularizer
            self._SupportVectorMachine = self._SupportVectorMachine.setRegParam(regularizer)
        else:
            raise ValueError("Invalid regularizer")

        # Indicate whether the GPUs will be used or not 
        if use_gpu in set([True, False]):
            self.__params['use_gpu'] = use_gpu
            self._SupportVectorMachine = self._SupportVectorMachine.setUseGpu(use_gpu)
        else:
            raise ValueError("Invalid use_gpu value: use_gpu must be set to either True or False")

        # Set the verbose parameter to indicate if the training loss will
        # be computed and printed at each epoch.
        if verbose in set([True,False]):
            self.__params['verbose'] = verbose
            self._SupportVectorMachine = self._SupportVectorMachine.setVerbose(verbose)
        else:
            raise ValueError("Invalid verbose value: must be set to a boolean value")

        if tol >=0:
            self.__params['tol'] = tol
            self._SupportVectorMachine = self._SupportVectorMachine.setTol(tol)
        else:
            raise ValueError("Parameter tol: must be greater than or equal to 0")

        if return_training_history in set([None, 'summary', 'full']):
            self.__params['return_training_history'] = return_training_history
            if(return_training_history is None):
                return_training_code = 0
            elif(return_training_history is 'summary'):
                return_training_code = 1
            else:
                return_training_code = 2
            self._SupportVectorMachine = self._SupportVectorMachine.setReturnTrainingCode(return_training_code)
        else:
            raise ValueError("Parameter return_training_history must be None, \'summary\' or \'full\'")

    def fit(self, data):
        """
        learn model

        :param data: data to fit model
        :type data: pyspark.sql.DataFrame/py4j.java_gateway.JavaObject which points to a memory address where the actual data is stored and it cannot be accessed by python as a python array.
        :returns: double -- final training loss of the last epoch
        """
        label_features = [ "label", "features" ]
        if all(x in label_features for x,y in data.dtypes):
            data = DatasetReader()._loadDF(data, 1)
        elif not 'ptr' in (x for x,y in data.dtypes):
            data = DatasetReader()._loadDF(data, 0, self.__params['labelColIndex'])

        history = self._SupportVectorMachine.fit(data._jdf)

        self.training_history_ = None
        if self.__params['return_training_history'] is not None:
            self.training_history_ = []
            for history_i in history:
                training_history = {}
                training_history['epochs'] = history_i.getEpochs();
                training_history['t_elap_sec'] = history_i.getTotTime();
                training_history['train_obj'] = history_i.getTrainCost();

                # don't output lists for summary
                for key in training_history:
                    if(len(training_history[key])==1):
                        training_history[key] = training_history[key][0]

                self.training_history_.append(training_history)

        # Assign the model to the coef_ attribute
        self.coef_ = []
        coef_list = self._SupportVectorMachine.getWeights()
        for coef in coef_list:
            self.coef_.append(np.frombuffer(coef, dtype=np.float64))
        if len(self.coef_) == 1:
            self.coef_ = self.coef_[0]

        return self.training_history_


    def predict(self, data, num_threads=0):
        """
        Predict label

        :param data: data to make predictions
        :type data: pyspark.sql.DataFrame/py4j.java_gateway.JavaObject which points to a memory address where the actual data is stored and cannot be accessed by python as an array but only can passed as a parameter to this function in order to get the predictions
        :param num_threads: the number of threads to use for inference (default 0 means use all avaliable threads)   
        :returns: a pointer which points to a com.ibm.snap.ml.DatasetWithPredictions java object. This pointer cannot be accessed by python but the user can access the predictions from the pred_array_ field which is a python array.
        """
        if not hasattr(self, "coef_") or np.any(self.coef_ == None):
            raise ValueError("This instance is not fitted yet")
        label_features = [ "label", "features" ]
        if all(x in label_features for x,y in data.dtypes):
            data = DatasetReader()._loadDF(data, 1)
        elif not 'ptr' in (x for x,y in data.dtypes):
            data = DatasetReader()._loadDF(data, 0)

        pred = self._SupportVectorMachine.predict(data._jdf, num_threads)

        self.pred_array_ = np.frombuffer(pred.getPredictionsArray(), dtype=np.float64)
        
        return pred

    def get_params():
        """
        :returns: all the initialized parameters of the Support Vector Machine Model as a python dictionary
        """
        return self.__params
