# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2018. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************


from __future__ import print_function
import time
import argparse
import os

## SPARK config

from pyspark import SparkConf, SparkContext
from pyspark.sql.session import SparkSession

defaultPath = os.getcwd()
CLI=argparse.ArgumentParser()
CLI.add_argument(
    "--use_gpu",
    action = 'store_true',
    default=False
)
CLI.add_argument(
    "--skip_sparkml",
    action = 'store_true',
    default=False
)
CLI.add_argument(
   "--data_path",
   type=str,
   default=defaultPath
)
CLI.add_argument(
   "--num_partitions",
   type=int,
   default= 4
)
CLI.add_argument(
   "--use_balanced_class_weights",
   action = 'store_true',
   default=False
)
args = CLI.parse_args()

if args.use_balanced_class_weights:
  class_weights = "balanced"
else:
  class_weights = None
# Set your own master (e.g 'spark://node.com:7077') or use 'local'
# Configure Spark (2 executors, local standalone mode)
# If one wants to run spark on a cluster, use your own spark configuration (set master , appName, etc...) 
conf = SparkConf()\
.set("spark.default.parallelism",args.num_partitions)\
.set("spark.cores.max",args.num_partitions)\
.setAppName("Spark-news20-SparkComparison")\
.set("spark.executor.extraJavaOptions","-XX:MaxDirectMemorySize=8192m")\
.set("spark.driver.maxResultSize",8192)
# Create spark config
sc = SparkContext(conf=conf)
# Configure Spark
spark = SparkSession(sc)

# disable spark INFO logging - comment out to enable
log4j = sc._jvm.org.apache.log4j
log4j.LogManager.getRootLogger().setLevel(log4j.Level.ERROR)


# data paths
# Please use the absolute path of the file if you wish to run the example on a distributed mode
train_filename = args.data_path + '/data/news20.binary-train.libsvm'
test_filename = args.data_path + '/data/news20.binary-test.libsvm'

## snapML

from snap_ml_spark import DatasetReader
from snap_ml_spark import LogisticRegression as snapml_LogisticRegression
from snap_ml_spark.Metrics import accuracy

n_features_ = 1355191


# Load training data
trainData = DatasetReader().setFormat("libsvm").setNumFt(n_features_).load(train_filename)

# Load test data
testData = DatasetReader().setFormat("libsvm").setNumFt(n_features_).load(test_filename)

# Create snapML Logistic Regression
snapml_regularizer = 100.0
snapml_lr = snapml_LogisticRegression(max_iter=10, regularizer=snapml_regularizer, verbose=False, dual=True, use_gpu=args.use_gpu, class_weights=class_weights)

# Fit the model and time it
snapml_t0 = time.time()
snapml_lr.fit(trainData)
snapml_time = time.time() - snapml_t0

# Perform inference on test data
pred = snapml_lr.predict(testData)

# Compute accuracy
snapml_accuracy = accuracy(pred)

# Print off SnapML result
print('snapML accuracy: %.4f' %snapml_accuracy, ", time: %.2f" % snapml_time, 's')


## SPARK MLlib

# Skip if --skip_sparkml flag is given
if args.skip_sparkml:
    print('Execution using Spark ML is skipped.')
    exit()

from pyspark.ml.classification import LogisticRegression as sparkmllib_LogisticRegression

train_filename = "file://" + train_filename
test_filename = "file://" + test_filename

# Load training data
trainData = spark.read.format("libsvm").option("numFeatures", str(n_features_)).load(train_filename) # for csv need to create createDataFrame(data, ["label", "features"])
testData = spark.read.format("libsvm").option("numFeatures", str(n_features_)).load(test_filename)

n_examples = trainData.count()
# print('number of examples:', n_examples)

# Create sparkML lib Logistic Regression
sparkmllib_lr = sparkmllib_LogisticRegression(fitIntercept=False, standardization=False, regParam=snapml_regularizer/n_examples)

# Fit the model and time it
sparkmllib_t0 = time.time()
sparkmllib_lr_model = sparkmllib_lr.fit(trainData)
sparkmllib_time = time.time() - sparkmllib_t0

# Perform inference on test data
predictions = sparkmllib_lr_model.transform(testData)

# Show predictions against test labels
# predictions.select("rawPrediction", "prediction", "label", "features").show(10)

# Compute accuracy
from pyspark.ml.evaluation import MulticlassClassificationEvaluator
evaluator = MulticlassClassificationEvaluator(labelCol="label", predictionCol="prediction", metricName="accuracy")
sparkmllib_accuracy = evaluator.evaluate(predictions)

# Print off Spark result
print('sparkML lib accuracy: %.4f' % sparkmllib_accuracy, ", time: %.2f" % sparkmllib_time, 's')
